<?php

/* markup.php - SynApp2 markup language generator
**
   +----------------------------------------------------------------------+
   | SynApp2 Version 1                                                    |
   +----------------------------------------------------------------------+
   | Copyright (c) 2007 - 2009 Richard Howell. All rights reserved.       |
   +----------------------------------------------------------------------+
   | This source file is subject to version 1.01 of the SynApp2 license,  |
   | that is bundled with this package in the file LICENSE, and is        |
   | available through the world-wide-web at the following url:           |
   | http://www.synapp2.org/license/1_01.txt                              |
   | If you did not receive a copy of the SynApp2 license and are unable  |
   | to obtain it through the world-wide-web, please send a note to       |
   | license@synapp2.org so we can mail you a copy immediately.           |
   +----------------------------------------------------------------------+
   | Authors: Richard Howell                                              |
   |                                                                      |
   +----------------------------------------------------------------------+
**
** http://www.synapp2.org
**
** $Id: markup.php,v 1.11 2009/05/01 20:40:34 richard Exp $
*/

class markup
{
    var $m_action = null;
    var $m_dbx = null;
    var $m_custom = null;
    var $m_schema = null;

    var $m_CSS = array(); // view style elements
    var $m_VML = array(); // view markup language elements
    var $m_RPT = array(); // view markup report filter/constraint elements (fform)
    var $m_IML = array(); // view markup language elements (iform)
    var $m_TML = array(); // view markup language elements (tform)
    var $m_XCH = array(); // view exchange objects
    var $m_MAP = array(); // view element id maps
    var $m_OVL = array(); // view form overlay elements
    var $m_ADH = array(); // adhoc elements
    var $m_IDD = array(); // drag 'n drop id's

    var $m_fk_info = array();
    var $m_caption_info = array();

    var $m_is_remote = false;

    function markup(&$action, &$dbx, &$custom, &$schema, $active = false)
    {
        $this->m_action = &$action;
        $this->m_dbx = &$dbx;
        $this->m_custom = &$custom;
        $this->m_schema = &$schema;

        if ($active)
        {
            $this->build();
        }
    }

    function build()
    {
        $is_not_synapp2 = strtolower($this->m_action->get_appid()) != SYNAPP2;

        if (!$this->m_action->is_adhoc())
        {
            return; // TODO: HACK: KLUGE:
        }
        else if ($this->m_action->get_qid() == '_app_nav_')
        {
            $this->app_nav_gen($this->m_action->get_appid(),
                               $this->m_action->get_pid(),
                               $this->m_action->get_rid(),
                               $this->m_action->get_col_var('document_pathname'),
                               $this->m_schema->get_table_names());
        }
        else if (($this->m_action->get_qid() == '_pagegen_show_' ||
                  $this->m_action->get_qid() == '_pagegen_process_') &&
                 $is_not_synapp2)
        {
            $activity = null; // set up for reflect_activity()

            ////

            $appid = $this->m_custom->get_database(); // TODO: customize $appid

            $basis_table_name = $this->m_action->get_col_var('basis_table');

            $paths = $this->m_schema->get_paths($basis_table_name);

            $this->erd_gen($this->m_custom->get_database(), $basis_table_name, $this->m_schema->get_table_names(), $paths, $this->m_ADH);

            if ($this->m_action->get_qid() == '_pagegen_process_')
            {
                $this->m_is_remote = $this->m_action->get_col_var('remote') == 'on';

                if (!empty($basis_table_name) && !empty($paths))
                {
                    $active_paths = null;

                    ////

                    $col_count = $this->m_schema->get_table_col_count($basis_table_name);

                    for ($basis_index = 0; $basis_index < $col_count; $basis_index++)
                    {
                        $basis_key = '_basis_key_' . sprintf('%02d', $basis_index);

                        // detect var by fishing with generated name

                        if (($col_var = $this->m_action->get_col_var($basis_key)) !== null)
                        {
                            list($path_index, $tier_table) = explode(',', $col_var);
                            $active_paths[$path_index] = $tier_table;
                        }
                    }

                    ////

                    $objective = $basis_table_name;

                    if (!empty($active_paths))
                    {
                        $ipath = 0;

                        foreach ($active_paths as $path_index => $tier_table)
                        {
                            $path_nodes = &$paths[$path_index][PATH];

                            $inode = 0;

                            foreach ($path_nodes as $path_node)
                            {
                                if ($path_node[TABLE_NAME] == $tier_table)
                                {
                                    break;
                                }

                                $fk_full_name = $path_node[TABLE_NAME] . '.' . $path_node[COL_NAME];

                                $is_prime = $ipath == 0;
                                $is_basis = $inode == 0;
                                $is_indepedent = $is_basis && !$is_prime;
                                $is_constraint = !$is_indepedent;

                                $this->m_fk_info[$fk_full_name] = array('is_prime'=>$is_prime,
                                                                        'is_basis'=>$is_basis,
                                                                        'is_constraint'=>$is_constraint,
                                                                        'is_independent'=>$is_indepedent,
                                                                        'nav'=>$ipath);

                                $inode++;
                            }

                            $ipath++;
                        }

                        ////

                        $ipath = 0;

                        foreach ($active_paths as $path_index => $tier_table)
                        {
                            $path_nodes = &$paths[$path_index][PATH];

                            if (!$this->m_is_remote)
                            {
                                $objective .= ',' . $path_nodes[0][COL_NAME] . ',' . $tier_table;
                            }

                            $this->gather_page_markup($tier_table, $basis_table_name, $path_nodes, $ipath, $ipath + 1 == count($active_paths));
                            $ipath++;
                        }
                    }
                    else
                    {
                        $path_nodes = array(array(TABLE_NAME=>$basis_table_name, COL_NAME=>null, JOIN_TABLE=>null, JOIN_COL=>null));
                        $this->gather_page_markup('', $basis_table_name, $path_nodes);
                    }

                    ////

                    for ($nav = 0; $nav < (empty($active_paths) ? 1 : count($active_paths)); $nav++)
                    {
                        $this->m_XCH[] = "put_cxlref($nav, new cxl('$basis_table_name', '$objective'));";
                    }

                    ////

                    if (!empty($this->m_caption_info))
                    {
                        foreach ($this->m_caption_info as $caption_key => $info)
                        {
                            if (isset($this->m_VML[$caption_key]))
                            {
                                $this->m_VML[$caption_key] = str_replace('<!--{caption_info}-->', " ({$info})", $this->m_VML[$caption_key]);
                            }
                        }
                    }
                }

                ////

                if (!empty($appid))
                {
                    $basis_page_outfile = $this->m_action->get_col_var('basis_page_outfile');
                    $basis_page_template = $this->m_action->get_col_var('basis_page_template');
                    $basis_page_enable = $this->m_action->get_col_var('basis_page_enable') == 'on';

                    if (!empty($basis_page_outfile) && !empty($basis_page_template) && $basis_page_enable)
                    {
                        $basis_page_result = $this->page_gen($appid, $basis_page_outfile, $basis_page_template);

                        if ($basis_page_result)
                        {
                            $components = explode('/', $_SERVER['PHP_SELF']);

                            if (!empty($components))
                            {
                                $href = str_replace(array_pop($components), $basis_page_result, $_SERVER['PHP_SELF']);
                                $activity[] = "Basis Page: <a href=\"{$href}\">{$basis_page_outfile}</a> (" . date ("Y-m-d H:i:s", filemtime($basis_page_result)) . ")";
                            }
                        }
                    }

                    ////

                    $report_page_outfile = $this->m_action->get_col_var('report_page_outfile');
                    $report_page_template = $this->m_action->get_col_var('report_page_template');
                    $report_page_enable = $this->m_action->get_col_var('report_page_enable') == 'on';

                    if (!empty($report_page_outfile) && !empty($report_page_template) && $report_page_enable)
                    {
                        $report_page_result = $this->page_gen($appid, $report_page_outfile, $report_page_template);

                        if ($report_page_result)
                        {
                            $components = explode('/', $_SERVER['PHP_SELF']);

                            if (!empty($components))
                            {
                                $reportid = str_replace('.htm', '', str_replace('.report', '', $report_page_outfile));
                                $href = str_replace(array_pop($components), "reports.htm?appid={$appid}&amp;reportid={$reportid}", $_SERVER['PHP_SELF']);
                                $activity[] = "Report Page: <a href=\"{$href}\">{$report_page_outfile}</a> (" . date ("Y-m-d H:i:s", filemtime($report_page_result)) . ")";
                            }
                        }
                    }

                    ////

                    // HACK: KLUGE: we get away with reusing the dirty markup instance because the templates employ only a few specific output tokens
                    $this->page_gen($appid, $appid, 'template.index.html');
                    $this->m_XCH = array(); // reset this markup member while others still have data!
                    $this->m_XCH[] = "g_appid = \"{$appid}\";";
                    $this->m_XCH[] = "put_cxlref(0, new cxl('welcome'));";
                    $this->page_gen($appid, $appid, 'template.welcome.htm');
                }
            }

            ////

            $this->reflect_activity($activity, $this->m_ADH);
        }
        else if (($this->m_action->get_qid() == '_keymap_show_' ||
                  $this->m_action->get_qid() == '_keymap_process_') &&
                 $is_not_synapp2)
        {
            if ($_fkey_ = $this->m_action->get_col_var('_fkey_'))
            {
                list($table_name, $col_name) = explode('.', $_fkey_);

                if (($_jkey_ = $this->m_action->get_col_var("_jkey_$table_name")) !== false)
                {
                    assert('!empty($table_name)');
                    assert('!empty($col_name)');

                    if (!empty($table_name) && !empty($table_name))
                    {
                        $affected_rows = 0;

                        if ($_jkey_ == 'delete')
                        {
                            if ($this->m_schema->is_table_present(KEYMAP))
                            {
                                $q = $this->m_dbx->get_keymap_table_query("delete from " . KEYMAP . " where table_name='$table_name' and col_name='$col_name'");
                                $this->m_dbx->query($q);
                                $affected_rows = $this->m_dbx->affected_rows();
                            }
                        }
                        else
                        {
                            list($join_table, $join_col) = explode('.', $_jkey_);

                            assert('!empty($join_table)');
                            assert('!empty($join_col)');

                            if (!empty($join_table) && !empty($join_col))
                            {
                                if (!$this->m_schema->is_table_present(KEYMAP))
                                {
                                    $this->m_dbx->create_keymap_table();
                                }
                                else
                                {
                                    $q = $this->m_dbx->get_keymap_table_query("update " . KEYMAP . " set join_table='$join_table', join_col='$join_col' where table_name='$table_name' and col_name='$col_name'");
                                    $this->m_dbx->query($q);
                                    $affected_rows = $this->m_dbx->affected_rows();
                                }

                                if ($affected_rows < 1)
                                {
                                    $q = $this->m_dbx->get_keymap_table_query("insert into " . KEYMAP . " (table_name, col_name, join_table, join_col) values ('$table_name', '$col_name', '$join_table', '$join_col')");
                                    $this->m_dbx->query($q);
                                    $affected_rows = $this->m_dbx->affected_rows();
                                    assert('$affected_rows == 1');
                                }
                            }
                        }
                    }
                }
            }

            $keys_info = $this->m_schema->get_keys_info();

            $this->fkey_gen($this->m_custom->get_database(), $keys_info, $this->m_ADH);
        }
    }

    function app_nav_gen($appid, $pid, $rid, $document_pathname, $table_names)
    {
        if (!empty($appid) && !empty($pid) && !empty($document_pathname))
        {
            $appdir = "../{$appid}/"; // TODO: configurable
            $shared = "../_shared_/"; // TODO: configurable
            $file_ext = '.htm'; // TODO: configurable

            $tabs = array_merge($table_names, $this->m_custom->get_nav_incl_tabs());

            $active_tab_name = basename($document_pathname);

            if (!empty($tabs))
            {
                $lines_nav = array();

                $lines_nav[] = "<ul id=\"nav\">";

                foreach ($tabs as $tab_info)
                {
                    $tab_href = is_array($tab_info) ? $tab_info[A_HREF] : $tab_info ;
                    $tab_text = is_array($tab_info) ? $tab_info[A_TEXT] : trim(ucwords(str_replace('_', ' ', $tab_href))) ;

                    if ($this->m_schema->is_sys_table($tab_href) ||
                        $this->m_custom->is_nav_omit_tab($tab_href, $appid) ||
                        !$this->m_custom->is_auth_pid_username($tab_href, login_username()) ||
                        !file_exists("{$appdir}{$tab_href}{$file_ext}"))
                    {
                        continue;
                    }

                    $class_nav_state = ($tab_href . $file_ext == $active_tab_name) ? ' class="class_nav_active"' : ' class="class_nav_inactive"';

                    $lines_nav[] = "<li id=\"id__nav__{$tab_href}\"><a href=\"{$appdir}{$tab_href}{$file_ext}\"{$class_nav_state}>{$tab_text}</a></li>";
                }

                if ($this->m_custom->is_auth_pid_username('reports', login_username()))
                {
                    $class_nav_state = ($pid == 'reports') ? ' class="class_nav_active"' : ' class="class_nav_inactive"';

                    $lines_nav[] = "<li id=\"id__nav__reports\"><a href=\"{$shared}reports{$file_ext}?appid={$appid}\"{$class_nav_state}>Reports</a></li>";
                }

                $lines_nav[] = "<li id=\"id__nav__logout\"><a href=\"{$shared}logout.htm?appid={$appid}\" class=\"class_nav_inactive\">Logout</a></li>";

                $lines_nav[] = "</ul>";

                foreach ($lines_nav as $line)
                {
                    $this->m_ADH[] = $line;
                }

                ////

                if ($pid == 'reports')
                {
                    $lines_rpt = array();

                    $rpts = array_merge($table_names, $this->m_custom->get_rpt_incl_tabs());

                    if (!empty($rpts))
                    {
                        $lines_rpt[] = "<ul id=\"rpt\">";

                        foreach ($rpts as $tab_name)
                        {
                            if ($this->m_schema->is_sys_table($tab_name) ||
                                $this->m_custom->is_rpt_omit_tab($tab_name, $appid) ||
                                !$this->m_custom->is_auth_rid_username($tab_name, login_username()) ||
                                !file_exists("{$appdir}{$tab_name}.report{$file_ext}"))
                            {
                                continue;
                            }

                            $class_rpt_state = ($rid == $tab_name) ? ' class="class_rpt_active"' : ' class="class_rpt_inactive"';

                            $tab_text = trim(ucwords(str_replace('_', ' ', $tab_name)));

                            $lines_rpt[] = "<li id=\"id__rpt__{$tab_name}\"><a href=\"{$shared}reports{$file_ext}?appid={$appid}&amp;reportid={$tab_name}\"{$class_rpt_state}>{$tab_text}</a></li>";
                        }

                        $lines_rpt[] = "</ul>";
                    }

                    foreach ($lines_rpt as $line)
                    {
                        $this->m_ADH[] = $line;
                    }
                }
            }
        }
    }

    function erd_gen($database, $basis_table_name, &$table_names, &$paths, &$elements)
    {
        // NOTE: markup must be well-formed to return via xml

        $elements[] = "<div class=\"class_nav_col\">";

        $elements[] = "<div id=\"id_pagegen_erd\" class=\"class_layout_group_std\">";
        $elements[] = "<form id=\"id_pagegen_erd_form\" action=\"\" method=\"post\">";
        $elements[] = "<div class=\"class_form_std\">";

        $elements[] = "Database: $database";

        $elements[] = "<br />";
        $elements[] = "<br />";

        $buttons_enabled = false;

        if (!empty($table_names))
        {
            $elements[] = "<label>Basis Table:</label><select name=\"basis_table\" onchange=\"pagegen_show();\">";

            if (empty($paths))
            {
                $basis_table_name = null;
            }

            if (empty($basis_table_name))
            {
                $elements[] = "<option value=\"\" selected=\"selected\" /> -- choose -- ";
            }

            foreach ($table_names as $table_name)
            {
                if (!$this->m_schema->is_sys_table($table_name))
                {
                    $selected = $table_name == $basis_table_name ? " selected=\"selected\"" : '' ;
                    $elements[] = "<option value=\"$table_name\"$selected />" . $table_name;
                }
            }

            $elements[] = "</select><br />";

            if (empty($basis_table_name))
            {
                $elements[] = "</div><!-- class_form_std -->";
            }
            else
            {
                $basis_page_outfile = $basis_table_name . '.htm';
                $report_page_outfile = $basis_table_name . '.report.htm';

                $elements[] = "<label>Basis Page:</label><input type=\"text\" name=\"basis_page_outfile\" value=\"{$basis_page_outfile}\" /><br />";
                $elements[] = "<label>Basis Page Template:</label><input type=\"text\" name=\"basis_page_template\" value=\"template.htm\" />";
                $elements[] = "<input type=\"checkbox\" name=\"basis_page_enable\" title=\"Enable basis page markup generation\" checked=\"checked\" /><br />";
//                $elements[] = "<label>Remote:</label><input type=\"checkbox\" name=\"remote\" /><br />";
                $elements[] = "<label>Report Page:</label><input type=\"text\" name=\"report_page_outfile\" value=\"{$report_page_outfile}\" /><br />";
                $elements[] = "<label>Report Page Template:</label><input type=\"text\" name=\"report_page_template\" value=\"template.report.htm\" />";
                $elements[] = "<input type=\"checkbox\" name=\"report_page_enable\" title=\"Enable report page markup generation\" checked=\"checked\" /><br />";

                $elements[] = "</div><!-- class_form_std -->";

                $elements[] = "<br />";

                $elements[] = "<table class=\"class_table_std\" cellspacing=\"0\">";
                $elements[] = "<tbody>";

                ////

                $topo_paths = array();

                foreach ($paths as $path_index => $path)
                {
                    $topo_key = '';

                    foreach ($path[PATH] as $node_index => $path_node)
                    {
                        $table_name = $path_node[TABLE_NAME];
                        $col_name = $path_node[COL_NAME];
                        $topo_key .= $table_name . sprintf('%02d', $this->m_schema->get_col_index($table_name, $col_name)) . $col_name;
                    }

                    $topo_paths[$topo_key] = array(PATH=>$path[PATH], 'path_index'=>$path_index);
                }

                ksort($topo_paths);

                ////

                $tiers = array();
                $tier_cols = count($topo_paths);

                $tier_index = 0;
                $col_name = $this->m_custom->get_pk_name($basis_table_name);
                $basis_index = $this->m_schema->get_col_index($basis_table_name, $col_name);
                $basis_key = '_basis_key_' . sprintf('%02d', $basis_index);
                $node_key =  $col_name . $basis_key;

                $tiers[$tier_index][$node_key]['table'] = $basis_table_name;
                $tiers[$tier_index][$node_key]['colspan'] = $tier_cols;
                $tiers[$tier_index][$node_key]['align'] = 0;
                $tiers[$tier_index][$node_key]['path_index'] = null;
                $tiers[$tier_index][$node_key]['basis_index'] = $basis_index;

                $col_index = 0;

                foreach ($topo_paths as $path_info)
                {
                    $path = &$path_info[PATH];
                    $path_index = $path_info['path_index'];

                    $basis_index = $this->m_schema->get_col_index($basis_table_name, $path[0][COL_NAME]);
                    $basis_key = '_basis_key_' . sprintf('%02d', $basis_index);

                    foreach ($path as $node_index => $path_node)
                    {
                        $tier_index = $node_index + 1;

                        if (!empty($path_node[JOIN_TABLE]))
                        {
                            $node_key = $path_node[COL_NAME] . $basis_key;

                            if (!isset($tiers[$tier_index][$node_key]['colspan']))
                            {
                                $tiers[$tier_index][$node_key]['table'] = $path_node[JOIN_TABLE];
                                $tiers[$tier_index][$node_key]['colspan'] = 1;
                                $tiers[$tier_index][$node_key]['align'] = $col_index;
                                $tiers[$tier_index][$node_key]['path_index'] = $path_index;
                                $tiers[$tier_index][$node_key]['basis_index'] = $basis_index;
                            }
                            else
                            {
                                $tiers[$tier_index][$node_key]['colspan'] += 1;
                            }
                        }
                    }

                    $col_index++;
                }

                ////

                $table_tr = array();

                foreach ($tiers as $tier_index => $tier)
                {
                    $col_index = 0;

                    $tr = '<tr>';

                    foreach ($tier as $node_key => $tier_col)
                    {
                        list($col_name, $basis_key) = explode('_basis_key_', $node_key);

                        $table = $tier_col['table'];
                        $colspan = $tier_col['colspan'];
                        $align = $tier_col['align'];
                        $path_index = $tier_col['path_index'];
                        $basis_index = $tier_col['basis_index'];

                        while ($col_index < $align)
                        {
                            $tr .= '<td></td>';
                            $col_index++;
                        }

                        $input_radio = $tier_index ? "<br /><input type=\"radio\" name=\"_basis_key_$basis_key\" value=\"$path_index,$table\" />" : '' ;

                        $tr .= "<td colspan=\"$colspan\">$table<br />($col_name)$input_radio</td>";

                        $col_index += $colspan;
                    }

                    while ($col_index < $tier_cols)
                    {
                        $tr .= '<td></td>';
                        $col_index++;
                    }

                    array_unshift($table_tr, $tr . '</tr>');
                }

                ////

                if (!empty($table_tr))
                {
                    foreach ($table_tr as $tr)
                    {
                        $elements[] = $tr;
                    }

                    $buttons_enabled = true;
                }

                ////

                $elements[] = "</tbody>";
                $elements[] = "</table>";
            }
        }
        else
        {
            $elements[] = "<p>There is no information for this database.</p>";
            $elements[] = "</div><!-- class_form_std -->";
        }

        $disabled = $buttons_enabled ? '' : "disabled=\"disabled\" " ;

        $elements[] = "<div class=\"class_vsep_std\"></div>";
        $elements[] = "<div class=\"class_nav_controls\">";
        $elements[] = "<input {$disabled}type=\"button\" value=\"Go\" onclick=\"pagegen();\" />";
        $elements[] = "<input {$disabled}type=\"button\" value=\"Reset\" onclick=\"document.getElementById('id_pagegen_erd_form').reset();\" />";
        $elements[] = "</div><!-- class_nav_controls -->";
        $elements[] = "</form>";
        $elements[] = "</div><!-- class_layout_group_std -->";

        $elements[] = "<div class=\"class_layout_vsep_std\"></div>";
        $elements[] = "<div class=\"class_layout_hsep_std\"></div>";

        $elements[] = "</div><!-- nav_col -->";

        ////

        $elements['activity'] = "<!-- activity -->"; // armed - set up for reflect_activity()
    }

    function reflect_activity($activity, &$elements)
    {
        // NOTE: markup must be well-formed to return via xml

        if (!empty($activity) && is_array($activity) && isset($elements['activity']))
        {
            $elements['activity'] = "<div class=\"class_nav_col\">\n";
            $elements['activity'] .= "<div class=\"class_layout_group_std\">\n";

            foreach ($activity as $element)
            {
                $elements['activity'] .= "$element<br />\n";
            }

            $elements['activity'] .= "</div><!-- class_layout_group_std -->\n";
            $elements['activity'] .= "<div class=\"class_layout_vsep_std\"></div>\n";
            $elements['activity'] .= "<div class=\"class_layout_hsep_std\"></div>\n";
            $elements['activity'] .= "</div><!-- nav_col -->\n";
        }
    }

    function fkey_options($table_name, &$keys_info, &$elements)
    {
        // NOTE: markup must be well-formed to return via xml

        $elements[] = "<option value=\"\" />";

        $join_tables = array_keys($keys_info); // support nested iteration of $keys_info via reference (TDB: pass by value may be better??)

        foreach ($join_tables as $join_table)
        {
            if ($join_table != $table_name)
            {
                $table_key_info = $keys_info[$join_table]; // this, hopefully, leaves current($keys_info) unchanged

                foreach ($table_key_info as $key_info)
                {
                    if ($key_info['is_pk'] && $key_info[JOIN_TABLE] != $table_name)
                    {
                        $elements[] = "<option value=\"{$key_info[JOIN_TABLE]}.{$key_info[JOIN_COL]}\" />{$key_info[JOIN_TABLE]}.{$key_info[JOIN_COL]}";
                        break;
                    }
                }
            }
        }

        $elements[] = "<option value=\"delete\" /> -- delete -- ";
    }

    function fkey_gen($database, &$keys_info, &$elements)
    {
        // NOTE: markup must be well-formed to return via xml

        if (!empty($keys_info))
        {
            $elements[] = "<form class=\"class_keymap\" action=\"\">";
            $elements[] = "<div>";

            foreach ($keys_info as $table_name => $table_key_info)
            {
                $id = "id_table_$table_name";

                $this->m_IDD[] = $id;

                $elements[] = "<table id=\"id_table_$table_name\" cellspacing=\"0\">";
                $elements[] = "<thead>";
                $elements[] = "<tr><th class=\"class_keymap_table_name\" colspan=\"4\">$table_name</th></tr>";
                $elements[] = "<tr><th></th><th>Col Name</th><th>Join Table</th><th>Join Col</th></tr>";
                $elements[] = "</thead>";
                $elements[] = "<tbody>";

                $fk_count = 0;

                foreach ($table_key_info as $key_info)
                {
                    if ($key_info['is_pk'])
                    {
                        $key_tag = ' (PK)';
                        $elements[] = "<tr><td></td><td>{$key_info[COL_NAME]}$key_tag</td><td>{$key_info[JOIN_TABLE]}</td><td>{$key_info[JOIN_COL]}</td></tr>";
                    }
                    else
                    {
                        $key_tag = !empty($key_info[JOIN_TABLE]) ? ' (FK' . (++$fk_count) . ')' : '' ;
                        $elements[] = "<tr><td><input type=\"radio\" name=\"_fkey_\" value=\"{$key_info[TABLE_NAME]}.{$key_info[COL_NAME]}\" /></td><td>{$key_info[COL_NAME]}$key_tag</td><td>{$key_info[JOIN_TABLE]}</td><td>{$key_info[JOIN_COL]}</td></tr>";
                    }
                }

                $elements[] = "</tbody>";

                $elements[] = "<tfoot>";
                $elements[] = "<tr><th></th><th></th><th class=\"class_keymap_select\" colspan=\"2\">";
                $elements[] = "<select name=\"_jkey_$table_name\" onchange=\"submit_keymap();\">";

                $this->fkey_options($table_name, $keys_info, $elements); // NOTE: nested iteration of $keys_info via reference (for memory efficiency)

                $elements[] = "</select>";
                $elements[] = "</th></tr>";
                $elements[] = "</tfoot>";
                $elements[] = "</table>";
            }

            $elements[] = '<p />';
            $elements[] = "</div>";
            $elements[] = "</form>";
        }
        else
        {
            $elements[] = "<div class=\"class_layout_group_std\">";
            $elements[] = "Database: $database";
            $elements[] = "<br />";
            $elements[] = "<br />";
            $elements[] = "<p>There is no information for this database.</p>";
            $elements[] = "</div><!-- class_layout_group_std -->";
        }
    }

    function gather_page_markup($tier_table, $basis_table, &$path_nodes, $ipath = 0, $ilast=true)
    {
        $sep = "\n<!-- //// -->\n";

        if (!empty($path_nodes) && is_array($path_nodes))
        {
            $is_prime = $ipath == 0;
            $is_standalone_table = $is_prime && count($path_nodes) == 1;

            ////

            $this->m_VML[] = "<div class=\"class_nav_col\">";

            ////

            $skip_flag = !empty($tier_table);

            for ($j = count($path_nodes) - 1; $j >= 0; $j--)
            {
                $nav = sprintf("%02d", $ipath);
                $qid = $path_nodes[$j][TABLE_NAME];
                $navqid = "{$nav}__{$qid}";

                ////

                $is_tier_table = empty($tier_table) ? true : $qid == $tier_table;
                $is_key_table = !$is_prime && $j == 1;
                $is_constraint_table = $is_prime && $j == 1;
                $is_basis_table = $qid == $basis_table;

                $is_mid_table = !$is_basis_table && !$is_key_table && !$is_tier_table;

                ////

                if ($skip_flag)
                {
                    if ($this->m_is_remote)
                    {
                        if (!$is_basis_table)
                        {
                            continue;
                        }
                    }
                    else if ($is_tier_table)
                    {
                        $skip_flag = false;
                    }
                    else
                    {
                        continue;
                    }
                }

                ////

                $nav_flags = '';

                if ($is_prime)
                {
                    $nav_flags .= 'prime ';
                }

                if ($is_standalone_table)
                {
                    $nav_flags .= 'standalone ';
                }

                if ($is_tier_table)
                {
                    $nav_flags .= 'tier ';
                }

                if ($is_constraint_table)
                {
                    $nav_flags .= 'constraint ';
                }

                if ($is_key_table)
                {
                    $nav_flags .= 'key ';
                }

                if ($is_basis_table)
                {
                    $nav_flags .= 'basis ';
                }

                if ($is_mid_table)
                {
                    $nav_flags .= 'mid ';
                }

                $nav_flags = trim($nav_flags);

                ////

                if ($is_prime || (!$is_prime && !$is_basis_table))
                {
                    if (!$is_basis_table || $is_standalone_table)
                    {
                        $this->m_CSS[] = "#id__dform__$navqid";
                        $this->m_CSS[] = "{";

                        if ($is_tier_table)
                        {
                            $this->m_CSS[] = "    display:none;";
                        }
                        else
                        {
                            $this->m_CSS[] = "    display:block;";
                        }

                        $this->m_CSS[] = "}";
                        $this->m_CSS[] = "";

                        $this->m_CSS[] = "#id__sform__$navqid";
                        $this->m_CSS[] = "{";

                        if ($is_tier_table)
                        {
                            $this->m_CSS[] = "    display:block;";
                        }
                        else
                        {
                            $this->m_CSS[] = "    display:none;";
                        }

                        $this->m_CSS[] = "}";
                        $this->m_CSS[] = "";
                    }

                    $this->m_CSS[] = "#id__tform__$navqid";
                    $this->m_CSS[] = "{";

                    if ($is_basis_table || $is_key_table)
                    {
                        $this->m_CSS[] = "    display:block;";
                    }
                    else
                    {
                        $this->m_CSS[] = "    display:none;";
                        $this->m_CSS[] = "    position:absolute;";
                    }

                    $this->m_CSS[] = "}";
                    $this->m_CSS[] = "";

                    $this->m_CSS[] = "#id__iform__$navqid";
                    $this->m_CSS[] = "{";
                    $this->m_CSS[] = "    display:none;";
                    $this->m_CSS[] = "}";
                    $this->m_CSS[] = "";

                    ////

                    $this->gather_form_elements($this->m_IML, $qid, "id__iform__$navqid", $ipath, $is_basis_table, $is_key_table, $is_standalone_table);

                    if (!$is_basis_table || $is_standalone_table)
                    {
                        $this->m_VML[] = $sep;

                        $this->gather_form_elements($this->m_VML, $qid, "id__dform__$navqid", $ipath, $is_basis_table, $is_key_table, $is_standalone_table);
                        $this->gather_form_elements($this->m_VML, $qid, "id__sform__$navqid", $ipath, $is_basis_table, $is_key_table, $is_standalone_table);

                        if (!$is_basis_table)
                        {
                            $this->gather_form_elements($this->m_VML, $qid, "id__tform__$navqid", $ipath, $is_basis_table, $is_key_table, $is_standalone_table);
                        }
                        else
                        {
                            $this->gather_form_elements($this->m_TML, $qid, "id__tform__$navqid", $ipath, $is_basis_table, $is_key_table, $is_standalone_table);
                        }
                    }
                    else
                    {
                        $this->gather_form_elements($this->m_TML, $qid, "id__tform__$navqid", $ipath, $is_basis_table, $is_key_table, $is_standalone_table);
                    }

                    if ($is_basis_table)
                    {
                        $this->m_RPT[] = "";
                        $this->gather_form_elements($this->m_RPT, $qid, "id__fform__$navqid", $ipath, $is_basis_table, $is_key_table, $is_standalone_table);
                    }

                    $this->m_XCH['g_appid'] = "g_appid = \"" . $this->m_action->get_appid() . "\";";

                    $this->m_OVL[] = "reg_overlay(\"$navqid\", \"id__sform__$navqid\", \"search:$ipath:$qid\", \"$nav_flags\");";
                    $this->m_OVL[] = "reg_overlay(\"$navqid\", \"id__dform__$navqid\", \"display:$ipath:$qid\", \"$nav_flags\");";
                    $this->m_OVL[] = "reg_overlay(\"$navqid\", \"id__iform__$navqid\", \"input:$ipath:$qid\", \"$nav_flags\");";
                    $this->m_OVL[] = "reg_overlay(\"$navqid\", \"id__tform__$navqid\", \"select:$ipath:$qid\", \"$nav_flags\");";
                    $this->m_OVL[] = "reg_overlay(\"$navqid\", \"id__aform__$navqid\", \"stat:$ipath:$qid\", \"$nav_flags\");";
                }
            }

            ////

            $this->m_VML[] = "</div><!-- nav_col -->";

            ////

            if ($ilast && !empty($this->m_TML))
            {
                $this->m_VML[] = $sep;
                $this->m_VML[] = "<div class=\"class_layout_vsep_std\"></div>";
                $this->m_VML[] = "<div class=\"class_layout_hsep_std\"></div>";
                $this->m_VML[] = '';
                $this->m_VML[] = "<div id=\"id_basis_div\" class=\"class_nav_col\">";
                $this->m_VML = array_merge($this->m_VML, $this->m_TML);
                $this->m_VML[] = "</div><!-- id_basis_div -->";
            }

            ////

            if (!empty($this->m_VML))
            {
                $this->m_VML[] = $sep;
            }
        }
    }

    function gather_form_elements(&$vml, $qid, $id_pfx, $nav, $is_basis_table = false, $is_key_table = false, $is_standalone_table = false)
    {
        $sform = strpos($id_pfx, 'sform') !== false;
        $tform = strpos($id_pfx, 'tform') !== false;
        $dform = strpos($id_pfx, 'dform') !== false;
        $iform = strpos($id_pfx, 'iform') !== false;
        $fform = strpos($id_pfx, 'fform') !== false;

        $table_name = $this->m_custom->get_table($qid);

        $cols_info = $this->m_schema->get_table_cols_info($table_name);

        if (!empty($cols_info))
        {
            $ffb = array();
            $frm = array();
            $tbl = array();
            $th = '';
            $tbl_extra = array();
            $th_extra = '';
            $lookups = array();
            $lookup_cols = array();
            $tcl = ''; // ' /'; // tag closure - depends on !DOCTYPE // TODO: configurable
            $img = '../_shared_/img/'; // TODO: configurable
            $limit_rows = '5'; // TODO: configurable

            $navqid = sprintf("%02d__%s", $nav, $qid);
            $f1b = $is_basis_table ? '' : '1b' ;

            if ($tform)
            {
                $tbl[] = "\n<td style=\"display:none;\"></td>\n<td axis=\"_recno_\">&nbsp;</td>"; // setup for record counter
            }

            foreach ($cols_info as $col_info)
            {
                $col_name = $col_info['Field'];

                $is_pk = $this->m_schema->is_pk($table_name, $col_name);

                if (!$is_pk)
                {
                    $is_fk = $this->m_schema->is_fk($table_name, $col_name);
                    $is_auto = $this->m_schema->is_auto($table_name, $col_name);
                    $is_blind = $this->m_schema->is_blind($table_name, $col_name);

                    $full_name = $table_name . '.' . $col_name;

                    $is_constraint = !empty($this->m_fk_info[$full_name]) ? $this->m_fk_info[$full_name]['is_constraint'] : false ; // TODO: encapsulate
                    $is_independent = !empty($this->m_fk_info[$full_name]) ? $this->m_fk_info[$full_name]['is_independent'] : false ; // TODO: encapsulate

                    $col_display_name = $this->m_schema->get_col_display_name($col_info) ;
                    $col_type_detail = $this->m_schema->get_col_type_detail($col_info);

                    $col_id = "{$id_pfx}__{$col_name}";

                    ////

                    $flen = '';
                    $tlen = '';

                    $maxlength = $col_type_detail['maxlength'];

                    if ($col_type_detail['type'] == 'TEXT' || $maxlength > 40)
                    {
                        if ($sform || $dform || $fform)
                        {
                            $maxlength = 40; // input // TODO: configurable
                        }
                        else if ($iform )
                        {
                            $tlen = " cols=\"40\" rows=\"3\""; // textarea // TODO: configurable
                        }
                    }
                    else if (!$maxlength || $is_fk)
                    {
                        $maxlength = 20; // TODO: configurable
                    }

                    if (!$tlen)
                    {
                        $fsize = max(5, (int)($maxlength * 1.2)); // TODO: configurable
                        $flen = $iform ? " maxlength=\"$maxlength\" size=\"$fsize\"" : " size=\"$fsize\"";
                    }

                    ////

                    if (!$tform)
                    {
                        $element = '';

                        if ($iform && ($is_fk || $this->m_schema->is_set($table_name, $col_name)))
                        {
                            if (!$is_constraint && !$is_independent)
                            {
                                $element = "<select id=\"$col_id\" name=\"$col_name\" onchange=\"process_change(this);\"><option value=\"\"{$tcl}></select><br{$tcl}>";

                                if ($is_fk) // i.e. not is_set (not an enum)
                                {
                                    $fk_expand_qid = $this->m_custom->get_parent_name($table_name, $col_name);

                                    $lookups[] = array(JOIN_TABLE=>$fk_expand_qid,
                                                       JOIN_COL=>$col_name,
                                                       'reg_expr'=>"reg_lookup('{$id_pfx}', {qid:'{$fk_expand_qid}', basis_id:'{$col_id}'});");
                                }
                            }
                            else if ($is_independent)
                            {
                                $element = "<input readonly id=\"$col_id\" type=\"text\" name=\"{$table_name}_{$col_name}\"$flen{$tcl}><br{$tcl}>"; // TODO: TBD: '_'??

                                $this->m_MAP[] = "reg_blind('{$col_id}');"; // TODO: HACK: KLUGE: properly register independent key elements

                                $fk_expand_qid = $this->m_custom->get_parent_name($table_name, $col_name);
                                $fk_expand_select_id = sprintf("id__tform__%02d__%s", $this->m_fk_info[$full_name]['nav'], $fk_expand_qid);

                                $lookups[] = array(JOIN_TABLE=>$fk_expand_qid,
                                                   JOIN_COL=>$col_name,
                                                   'reg_expr'=>"reg_lookup('{$id_pfx}', {qid:'{$fk_expand_qid}', select_id:'{$fk_expand_select_id}'});");
                            }
                        }
                        else
                        {
                            $is_form_omit_col = ($dform && $this->m_custom->is_form_omit_col(DFORM, $col_name, $qid)) ||
                                                ($sform && $this->m_custom->is_form_omit_col(SFORM, $col_name, $qid)) ||
                                                ($fform && $this->m_custom->is_form_omit_col(FFORM, $col_name, $qid));

                            if ((!$is_constraint || $fform) && !$is_form_omit_col)
                            {
                                $readonly = $dform || ($iform && ($is_auto || $is_blind)) ? ' readonly' : '' ;
                                $input_name = !$is_fk ? $col_name : $table_name . "_$col_name" ;

                                if (empty($tlen))
                                {
                                    $element = "<input$readonly id=\"$col_id\" type=\"text\" name=\"$input_name\"$flen{$tcl}><br{$tcl}>";
                                }
                                else
                                {
                                    $element = "<textarea$readonly id=\"$col_id\" name=\"$input_name\"$tlen></textarea><br{$tcl}>";
                                }

                                if ($iform)
                                {
                                    if ($is_blind)
                                    {
                                        $this->m_MAP[] = "reg_blind('{$col_id}');";
                                    }
                                    else
                                    {
                                        $lookup_cols[] = $col_name;

                                        if ($col_type_detail['type'] == 'DATE')
                                        {
                                            $datepicker_format =  $this->m_dbx->get_datepicker_format();
                                            $this->m_MAP[] = "reg_date('{$col_id}', '{$datepicker_format}');";
                                        }
                                    }
                                }
                            }
                        }

                        if (!empty($element))
                        {
                            $frm[] = "<label for=\"$col_id\">{$col_display_name}:</label>";
                            $frm[] = $element;

                            if ($iform)
                            {
                                $ffb[] = "<label for=\"$col_id\"></label>";
                            }
                        }
                    }
                    else if ($tform)
                    {
                        $is_tform_omit_col = $this->m_custom->is_form_omit_col(TFORM, $col_name, $qid);

                        if (!$is_constraint && !$is_tform_omit_col)
                        {
                            $th .= "\n<th>$col_display_name</th>";
                            $axis_name = !$is_fk ? $col_name : "{$table_name}_{$col_name}" ; // TODO: TBD: '_'??
                            $tbl[] = "\n<td axis=\"$axis_name\">&nbsp;</td>";
                        }

                        if ($is_fk && isset($this->m_fk_info[$full_name]['nav']))
                        {
                            $parent_name = $this->m_custom->get_parent_name($table_name, $col_name);

                            if (strcasecmp($parent_name, $col_display_name)) // test != case insensitive
                            {
                                $parent_navqid = sprintf("%02d__%s", $this->m_fk_info[$full_name]['nav'], $parent_name);

                                $this->m_caption_info['dform' . $parent_navqid] = $col_display_name;
                                $this->m_caption_info['sform' . $parent_navqid] = $col_display_name;
                                $this->m_caption_info['tform' . $parent_navqid] = $col_display_name;
                                $this->m_caption_info['iform' . $parent_navqid] = $col_display_name;
                            }
                        }

                        if (!$is_tform_omit_col)
                        {
                            $xtra = $this->m_custom->get_extra($table_name . '.' . $col_name, $qid);

                            if (!empty($xtra))
                            {
                                foreach ($xtra as $col_name => $col_expr)
                                {
                                    $th_extra .= "\n<th>" . trim(ucwords(str_replace('_', ' ', $col_name))) . '</th>';
                                    $tbl_extra[] = "\n<td axis=\"$col_name\">&nbsp;</td>";
                                }
                            }
                        }
                    }
                }
            }
            //
            // end - foreach ($cols_info ...

            ////

            if (!empty($lookups) && !empty($lookup_cols))
            {
                foreach ($lookups as $details)
                {
                    $parent_cols_info = $this->m_schema->get_table_cols_info($details[JOIN_TABLE], $details[JOIN_COL]);

                    assert('!empty($parent_cols_info)');

                    if (!empty($parent_cols_info))
                    {
                        $matches = null;

                        foreach ($parent_cols_info as $parent_col_info)
                        {
                            foreach ($lookup_cols as $i => $lookup_col_name)
                            {
                                if ($parent_col_info['Field'] == $lookup_col_name)
                                {
                                    $matches[$lookup_col_name] = $i; // now we have a reason for the lookup
                                }
                            }
                        }

                        $is_lookup_needed = !empty($matches);

                        if ($is_lookup_needed)
                        {
                            foreach ($matches as $lookup_col_name => $i)
                            {
                                unset($lookup_cols[$i]);
                            }
                        }

                        $this->m_MAP[] = ($is_lookup_needed ? '' : '//') . $details['reg_expr'];
                    }
                }
            }

            ////

            if (!empty($tbl))
            {
                $subquery = $this->m_custom->get_subquery($table_name);

                if (!empty($subquery))
                {
                    foreach ($subquery as $col_name => $col_expr)
                    {
                        $th_extra .= "\n<th>" . trim(ucwords(str_replace('_', ' ', $col_name))) . '</th>';
                        $tbl_extra[] = "\n<td axis=\"$col_name\">&nbsp;</td>";
                    }
                }
            }

            if (!empty($tbl_extra))
            {
                $th .= $th_extra;
                $tbl = array_merge($tbl, $tbl_extra);
            }

            ////

            if (!empty($frm))
            {
                $vml[] = "<div id=\"{$id_pfx}\" class=\"class_layout_group_std\">";

                $id_form = "{$id_pfx}__form";

                if ($iform)
                {
                    $vml[] = "<form id=\"$id_form\" action=\"\" method=\"post\" onsubmit=\"do_exchange{$f1b}('id__iform__$navqid'); return false;\">";
                    $this->m_MAP[] = "reg_args('{$id_pfx}', {_qid_:'{$qid}', _nav_:'{$nav}'});";

                }
                else if ($sform)
                {
                    $vml[] = "<form id=\"$id_form\" action=\"\" method=\"post\" onsubmit=\"do_search('id__sform__$navqid', 'id__tform__$navqid'); return false;\">";
                    $this->m_MAP[] = "reg_args('{$id_pfx}', {_nav_:'{$nav}'});";
                }
                else if ($dform)
                {
                    $vml[] = "<form id=\"$id_form\" action=\"\">";
                    $this->m_MAP[] = "reg_args('{$id_pfx}', {_nav_:'{$nav}'});";
                }
                else if ($fform)
                {
                    $vml[] = "<form id=\"$id_form\" action=\"../_shared_/action.php\" method=\"post\">";
                }
                else
                {
                    $vml[] = "<form id=\"$id_form\" action=\"\">";
                }

                if (!empty($ffb))
                {
                    $vml[] = "<div id=\"{$id_pfx}_feedback\" class=\"class_form_std_feedback\">";
                    $vml = array_merge($vml, $ffb);
                    $vml[] = "</div><!-- class_form_std_feedback -->";
                }

                if (!empty($frm))
                {
                    $vml[] = "<div class=\"class_form_std\">";
                    $vml = array_merge($vml, $frm);
                    if ($fform)
                    {
                        $vml[] = "<input type=\"hidden\" name=\"_request_appid_\" value=\"" . $this->m_action->get_appid() . "\">";
                        $vml[] = "<input type=\"hidden\" name=\"_request_pid_\" value=\"reports\">"; // ASSUME: MAGIC: _pid_ for AUTH_PID
                        $vml[] = "<input type=\"hidden\" name=\"_request_qid_\" value=\"$qid\">";
                        $vml[] = "<input type=\"hidden\" name=\"_request_rid_\" value=\"$qid\">"; // ASSUME: MAGIC: _rid_ for AUTH_RID
                        $vml[] = "<input type=\"hidden\" name=\"_request_mode_\" value=\"report\">";
                        $vml[] = "<input type=\"hidden\" name=\"_request_show_response_\" value=\"0\">";
                    }
                    $vml[] = "</div><!-- class_form_std -->";
                }

                $vml[] = "</form><!-- $id_form -->";

                if ($sform)
                {
                    $vml[] = "<div class=\"class_vsep_std\"></div>";
                    $vml[] = "<div class=\"class_nav_controls\">";

                    $vml[] = "<div class=\"class_control_bar\">";
                    $vml['sform' . $navqid] = "<div class=\"class_control_bar_caption\">" . trim(ucwords(str_replace('_', ' ', $qid))) . "<!--{caption_info}--></div>";
                    $vml[] = "<div class=\"class_control_bar_controls\">";

                    $vml[] = "<img src=\"{$img}action_search.gif\" alt=\"Search\" title=\"Search\" onclick=\"do_search('id__sform__$navqid', 'id__tform__$navqid', this);\"{$tcl}>";
                    $vml[] = "<img src=\"{$img}action_reset.gif\" alt=\"Reset\" title=\"Reset\" onclick=\"do_reset('id__sform__$navqid', 'id__tform__$navqid', this);\"{$tcl}>";
                    $vml[] = "<img src=\"{$img}action_add.gif\" alt=\"Add\" title=\"Add\" onclick=\"do_add{$f1b}('id__tform__$navqid', 'id__iform__$navqid', this);\"{$tcl}>";

                    $vml[] = "</div><!-- class_control_bar_controls -->";
                    $vml[] = "<div class=\"class_control_bar_prompt\">Fill in values and then click Search</div>";
                    $vml[] = "</div><!-- class_control_bar -->";

                    $vml[] = "</div><!-- class_nav_controls -->";
                }
                else if ($iform)
                {
                    $vml[] = "<div class=\"class_vsep_std\"></div>";
                    $vml[] = "<div class=\"class_nav_controls\">";
                    $vml[] = "<input id=\"{$id_pfx}__ok\" value=\"Ok\" type=\"button\" onclick=\"do_exchange{$f1b}('id__iform__$navqid', this);\"{$tcl}>";
                    $vml[] = "<input id=\"{$id_pfx}__cancel\" value=\"Cancel\" type=\"button\" onclick=\"do_cancel{$f1b}('id__iform__$navqid', this);\"{$tcl}>";
                    $vml[] = "</div><!-- class_nav_controls -->";
                }
                else if ($fform)
                {
                    $vml[] = "<div class=\"class_vsep_std\"></div>";
                    $vml[] = "<div class=\"class_nav_controls\">";

                    $vml[] = "<div class=\"class_control_bar\">";
                    $vml[] = "<div class=\"class_control_bar_caption\">" . trim(ucwords(str_replace('_', ' ', $qid))) . "<!--{caption_info}--></div>";
                    $vml[] = "<div class=\"class_control_bar_controls\">";

                    $vml[] = "<img src=\"{$img}action_report.gif\" alt=\"Report\" title=\"Report\" onclick=\"document.getElementById('$id_form').submit();\"{$tcl}>";

                    $vml[] = "</div><!-- class_control_bar_controls -->";
                    $vml[] = "<div class=\"class_control_bar_prompt\">Fill in values and then click Report</div>";
                    $vml[] = "</div><!-- class_control_bar -->";

                    $vml[] = "</div><!-- class_nav_controls -->";
                }

                ////

                $vml[] = "</div><!-- $id_pfx -->";
            }
            else if (!empty($tbl))
            {
                $vml[] = "<div id=\"$id_pfx\" class=\"class_layout_group_std\">";

                ////

                $vml[] = "<table id=\"{$id_pfx}__table\"  class=\"class_table_std\" cellspacing=\"0\">";

                if (!empty($th))
                {
                    $vml[] = '<thead>';
                    $vml[] = "<tr>\n<th>&nbsp;</th>$th\n</tr>";
                    $vml[] = '</thead>';
                }

                $vml[] = "<tbody>";

                $tr = "<tr>";

                foreach ($tbl as $element)
                {
                    $tr .= $element;
                }

                $tr .= "\n</tr>";

                $vml[] = $tr;

                $vml[] = "</tbody>";
                $vml[] = "</table>";

                ////

                $this->m_MAP[] = "reg_args('{$id_pfx}', {_qid_:'{$qid}', _nav_:'{$nav}', _order_:'', _limit_rows_:'{$limit_rows}'});";

                ////

                $id_pfx_aform = "id__aform__$navqid";

                $vml[] = "<div class=\"class_vsep_std\"></div>";

                $vml[] = "<form id=\"{$id_pfx_aform}\" action=\"\">";
                $vml[] = "<div class=\"class_form_std\">";
                $vml[] = "<label for=\"{$id_pfx_aform}__found_rows\">Rows:</label>";
                $vml[] = "<input readonly id=\"{$id_pfx_aform}__found_rows\" type=\"text\" name=\"found_rows\" size=\"10\"{$tcl}><br{$tcl}>";
                $vml[] = "</div><!-- class_status_std -->";
                $vml[] = "</form><!-- {$id_pfx_aform} -->";

                $this->m_MAP[] = "reg_blind('{$id_pfx_aform}__found_rows');";

                ////

                $vml[] = "<div class=\"class_vsep_std\"></div>";

                $vml[] = "<div class=\"class_nav_controls\">";

                $vml[] = "<div class=\"class_control_bar\">";
                $vml['tform' . $navqid] = "<div class=\"class_control_bar_caption\">" . trim(ucwords(str_replace('_', ' ', $qid))) . "<!--{caption_info}--></div>";
                $vml[] = "<div class=\"class_control_bar_controls\">";

                $vml[] = "<img src=\"{$img}action_nav_first.gif\" alt=\"First\" title=\"First\" onclick=\"do_nav_first('id__tform__$navqid', null, this);\"{$tcl}>";
                $vml[] = "<img src=\"{$img}action_nav_up.gif\" alt=\"PgUp\" title=\"PgUp\" onclick=\"do_nav_up('id__tform__$navqid', this);\"{$tcl}>";
                $vml[] = "<img src=\"{$img}action_nav_dn.gif\" alt=\"PgDn\" title=\"PgDn\" onclick=\"do_nav_dn('id__tform__$navqid', this);\"{$tcl}>";
                $vml[] = "<img src=\"{$img}action_nav_last.gif\" alt=\"Last\" title=\"Last\" onclick=\"do_nav_last('id__tform__$navqid', this);\"{$tcl}>";

                $vml[] = "<img src=\"{$img}action_close.gif\" alt=\"Close\" title=\"Close\" onClick=\"do_nav_close('id__tform__$navqid', this);\"{$tcl}>";

                if ($is_basis_table)
                {
                    $vml[] = "<img src=\"{$img}action_add.gif\" alt=\"Add\" title=\"Add\" onclick=\"do_add('id__tform__$navqid', 'id__iform__$navqid', this);\"{$tcl}>";
                    $vml[] = "<img src=\"{$img}action_edit.gif\" alt=\"Edit\" title=\"Edit\" onclick=\"do_edit('id__tform__$navqid', 'id__iform__$navqid', this);\"{$tcl}>";
                    $vml[] = "<img src=\"{$img}action_remove.gif\" alt=\"Delete\" title=\"Delete\" onclick=\"do_delete('id__tform__$navqid', 'id__iform__$navqid', this);\"{$tcl}>";
                }
                else
                {
                    $vml[] = "<img src=\"{$img}action_add.gif\" alt=\"Add\" title=\"Add\" onclick=\"do_add{$f1b}('id__tform__$navqid', 'id__iform__$navqid', this);\"{$tcl}>";
                    $vml[] = "<img src=\"{$img}action_edit.gif\" alt=\"Edit\" title=\"Edit\" onclick=\"do_edit{$f1b}('id__tform__$navqid', 'id__iform__$navqid', this);\"{$tcl}>";
                    $vml[] = "<img src=\"{$img}action_remove.gif\" alt=\"Delete\" title=\"Delete\" onclick=\"do_delete{$f1b}('id__tform__$navqid', 'id__iform__$navqid', this);\"{$tcl}>";

                    if (!$is_key_table)
                    {
                        $vml[] = "<img src=\"{$img}action_nav_select.gif\" alt=\"Select\" title=\"Select\" onClick=\"do_nav_select('id__tform__$navqid', this);\"{$tcl}>";
                    }
                }

                $vml[] = "</div><!-- class_control_bar_controls -->";
                $vml[] = "<div class=\"class_control_bar_prompt\">Navigate records, select a record and then click Edit (or click Add to create a new record)</div>";
                $vml[] = "</div><!-- class_control_bar -->";

                $vml[] = "</div><!-- class_nav_controls -->";

                ////

                $vml[] = "</div><!-- $id_pfx -->";
            }

            if (!$iform && !$fform)
            {
                $vml[] = '';
                $vml[] = "<div class=\"class_layout_vsep_std\"></div>";
                $vml[] = "<div class=\"class_layout_hsep_std\"></div>";
            }

            $vml[] = '';
        }
    }

    function page_gen($appid, $filename, $template)
    {
        $result = '';

        $appdir = null;

        if (empty($appid) || empty($filename) || empty($template) || !($appdir = mkdir_app_dir($appid)))
        {
            return $result;
        }

        ////

        $title = array_shift(explode('.', $filename));

        if ($template == 'template.index.html' ||
            $template == 'template.welcome.htm')
        {
            $filename = str_replace('template.', '', $template);

            if (file_exists($appdir . $filename))
            {
                return $result;
            }
        }

        ////

        $input =  get_app_dir(SYNAPP2) . $template; // TODO: configurable template dir
        $output = $appdir . $filename;

        $lines_in = array();

        if ($lines_in = file($input))
        {
            $lines_out = array();

            $changes = false;
            $skip = false;

            $version_on = false;
            $title_on = false;
            $herald_on = false;

            $css_on = false;
            $vml_on = false;
            $rpt_on = false;
            $iml_on = false;
            $xch_on = false;
            $map_on = false;
            $ovl_on = false;
            $nav_on = false;

            foreach ($lines_in as $line)
            {
                if (strpos($line, '<!--{version}-->') === 0)
                {
                    $version_on = true;
                    $skip = true;
                }
                else if (strpos($line, '<!--{/version}-->') === 0)
                {
                    $skip = false;
                }
                else if (strpos($line, '<!--{title}-->') === 0)
                {
                    $title_on = true;
                    $skip = true;
                }
                else if (strpos($line, '<!--{/title}-->') === 0)
                {
                    $skip = false;
                }
                else if (strpos($line, '<!--{herald}-->') === 0)
                {
                    $herald_on = true;
                    $skip = true;
                }
                else if (strpos($line, '<!--{/herald}-->') === 0)
                {
                    $skip = false;
                }
                else if (strpos($line, '<!--{css}-->') === 0)
                {
                    $css_on = true;
                    $skip = true;
                }
                else if (strpos($line, '<!--{/css}-->') === 0)
                {
                    $skip = false;
                }
                else if (strpos($line, '<!--{iform}-->') === 0)
                {
                    $iml_on = true;
                    $skip = true;
                }
                else if (strpos($line, '<!--{/iform}-->') === 0)
                {
                    $skip = false;
                }
                else if (strpos($line, '<!--{group}-->') === 0)
                {
                    $vml_on = true;
                    $skip = true;
                }
                else if (strpos($line, '<!--{/group}-->') === 0)
                {
                    $skip = false;
                }
                else if (strpos($line, '<!--{rpt}-->') === 0)
                {
                    $rpt_on = true;
                    $skip = true;
                }
                else if (strpos($line, '<!--{/rpt}-->') === 0)
                {
                    $skip = false;
                }
                else if (strpos($line, '//<!--{xch}-->') === 0)
                {
                    $xch_on = true;
                    $skip = true;
                }
                else if (strpos($line, '//<!--{/xch}-->') === 0)
                {
                    $skip = false;
                }
                else if (strpos($line, '//<!--{map}-->') === 0)
                {
                    $map_on = true;
                    $skip = true;
                }
                else if (strpos($line, '//<!--{/map}-->') === 0)
                {
                    $skip = false;
                }
                else if (strpos($line, '//<!--{ovl}-->') === 0)
                {
                    $ovl_on = true;
                    $skip = true;
                }
                else if (strpos($line, '//<!--{/ovl}-->') === 0)
                {
                    $skip = false;
                }
                else if (strpos($line, '<!--{nav}-->') === 0)
                {
                    $nav_on = true;
                    $skip = true;
                }
                else if (strpos($line, '<!--{/nav}-->') === 0)
                {
                    $skip = false;
                }
                else if ($skip)
                {
                    continue;
                }

                $lines_out[] = $line;

                if ($version_on)
                {
                    $lines_out[] = '<!-- ' . get_version() . " -->\n";

                    $version_on = false;
                }
                else if ($title_on)
                {
                    $lines_out[] = "<title>\n";
                    $lines_out[] = ucfirst($title) . "\n";
                    $lines_out[] = "</title>\n";
                    $changes = true; // SAFETY: affect output ONLY if we actually insert markup

                    $title_on = false;
                }
                else if ($herald_on)
                {
                    $lines_out[] = ucfirst($title) . "\n";
                    $changes = true; // SAFETY: affect output ONLY if we actually insert markup

                    $herald_on = false;
                }
                else if ($css_on)
                {
                    if (!empty($this->m_CSS))
                    {
                        $lines_out[] = "<style type=\"text/css\">\n";

                        foreach ($this->m_CSS as $markup)
                        {
                            $lines_out[] = $markup . "\n";
                        }

                        $lines_out[] = "</style>\n";

                        $changes = true; // SAFETY: affect output ONLY if we actually insert markup
                    }

                    $css_on = false;
                }
                else if ($iml_on)
                {
                    foreach ($this->m_IML as $markup)
                    {
                        $lines_out[] = $markup . "\n";
                        $changes = true; // SAFETY: affect output ONLY if we actually insert markup
                    }

                    $iml_on = false;
                }
                else if ($vml_on)
                {
                    foreach ($this->m_VML as $markup)
                    {
                        $lines_out[] = $markup . "\n";
                        $changes = true; // SAFETY: affect output ONLY if we actually insert markup
                    }

                    $vml_on = false;
                }
                else if ($rpt_on)
                {
                    foreach ($this->m_RPT as $markup)
                    {
                        $lines_out[] = $markup . "\n";
                        $changes = true; // SAFETY: affect output ONLY if we actually insert markup
                    }

                    $rpt_on = false;
                }
                else if ($xch_on)
                {
                    foreach ($this->m_XCH as $markup)
                    {
                        $lines_out[] = $markup . "\n";
                        $changes = true; // SAFETY: affect output ONLY if we actually insert markup
                    }

                    $xch_on = false;
                }
                else if ($map_on)
                {
                    foreach ($this->m_MAP as $markup)
                    {
                        $lines_out[] = $markup . "\n";
                        $changes = true; // SAFETY: affect output ONLY if we actually insert markup
                    }

                    $map_on = false;
                }
                else if ($ovl_on)
                {
                    foreach ($this->m_OVL as $markup)
                    {
                        $lines_out[] = $markup . "\n";
                        $changes = true; // SAFETY: affect output ONLY if we actually insert markup
                    }

                    $ovl_on = false;
                }
                else if ($nav_on)
                {
                    foreach ($this->m_NAV as $markup)
                    {
                        $lines_out[] = $markup . "\n";
                        $changes = true; // SAFETY: affect output ONLY if we actually insert markup
                    }

                    $nav_on = false;
                }
            }

            if ($changes)
            {
                if (!writeFileContents($output, $lines_out))
                {
                    add_debug_msg("writeFileContents('$output') - failed"); // TODO: error message
                }

                $result = $output;
            }
            else
            {
                // TODO: output unaffected - message?
            }
        }

        return $result;
    }
}

////

?>
